# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, TypedDict

import pytest

from astrapy import AsyncCollection, AsyncDatabase, Collection, Database
from astrapy.constants import VectorMetric
from astrapy.info import CollectionDefinition
from astrapy.utils.unset import _UNSET

from ..conftest import (
    IS_ASTRA_DB,
    USE_RERANKER_API_KEY_HEADER,
    DefaultAsyncCollection,
    DefaultCollection,
)

RUN_RERANKER_TESTS = IS_ASTRA_DB or USE_RERANKER_API_KEY_HEADER


class MyTestDoc(TypedDict):
    p_bigint: int
    p_ascii: str


class MyTestMiniDoc(TypedDict):
    p_bigint: int


FIND_PROJECTION = {"_id": False, "p_bigint": True}
DOCUMENT = {"p_ascii": "abc", "p_bigint": 10000, "p_float": 0.123}
TYPED_DOCUMENT: MyTestDoc = {"p_ascii": "abc", "p_bigint": 10000}
FIND_FILTER = {"p_ascii": "abc", "p_bigint": 10000}
# find_and_rerank-related assets:
VLEX_DOCUMENT = {
    "p_ascii": "abc",
    "p_bigint": 10000,
    "p_float": 0.123,
    "$vector": [1, 2],
    "$lexical": "blo",
}
TYPED_VLEX_DOCUMENT: MyTestDoc = {
    "p_ascii": "abc",
    "p_bigint": 10000,
    "$vector": [1, 2],
    "$lexical": "blo",
}  # type: ignore[typeddict-unknown-key]


class TestCollectionTyping:
    @pytest.mark.describe("test of typing create_collection, sync")
    def test_create_collection_typing_sync(
        self,
        sync_database: Database,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        """Test of creating typed collections with generics (and not), sync."""

        # Untyped baseline
        c_co_untyped = sync_database.create_collection(
            sync_empty_collection.name,
            definition=(
                CollectionDefinition.builder()
                .set_vector_dimension(2)
                .set_vector_metric(VectorMetric.COSINE)
                .set_indexing("deny", ["not_indexed"])
                .build()
            ),
        )
        c_co_untyped.insert_one(DOCUMENT)
        cu_doc = c_co_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"_id", "p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_co_untyped_cursor = c_co_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = c_co_untyped_cursor.__next__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        c_co_untyped.delete_many({})

        # Typed
        c_co_typed: Collection[MyTestDoc] = sync_database.create_collection(
            sync_empty_collection.name,
            definition=(
                CollectionDefinition.builder()
                .set_vector_dimension(2)
                .set_vector_metric(VectorMetric.COSINE)
                .set_indexing("deny", ["not_indexed"])
                .build()
            ),
            document_type=MyTestDoc,
        )
        c_co_typed.insert_one(TYPED_DOCUMENT)
        ct_doc = c_co_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"_id", "p_ascii", "p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_co_typed_cursor = c_co_typed.find(
            {}, projection=FIND_PROJECTION, document_type=MyTestMiniDoc
        )
        ctcur_doc = c_co_typed_cursor.__next__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        c_co_typed.delete_many({})

    @pytest.mark.describe("test of typing get_collection, sync")
    def test_get_collection_typing_sync(
        self,
        sync_database: Database,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        """Test of getting typed collections with generics (and not), sync."""

        # Untyped baseline
        g_co_untyped = sync_database.get_collection(sync_empty_collection.name)
        g_co_untyped.insert_one(DOCUMENT)
        gu_doc = g_co_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"_id", "p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        g_co_untyped.delete_many({})

        # Typed
        g_co_typed: Collection[MyTestDoc] = sync_database.get_collection(
            sync_empty_collection.name,
            document_type=MyTestDoc,
        )
        g_co_typed.insert_one(TYPED_DOCUMENT)
        gt_doc = g_co_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"_id", "p_ascii", "p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        g_co_typed.delete_many({})

    @pytest.mark.describe("test of typing collection cursors with map, sync")
    def test_collection_cursormap_typing_sync(
        self,
        sync_database: Database,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        """Test of getting typed collections with generics (and not), sync."""

        g_co_untyped = sync_database.get_collection(sync_empty_collection.name)
        g_co_typed: Collection[MyTestDoc] = sync_database.get_collection(
            sync_empty_collection.name,
            document_type=MyTestDoc,
        )
        g_co_untyped.insert_many([DOCUMENT] * 30)

        # base cursors, read from them (un/typed)
        u_cur0 = g_co_untyped.find()
        t_cur0 = g_co_typed.find()
        gu_doc = next(u_cur0)
        gt_doc = next(t_cur0)
        assert gu_doc is not None
        assert gt_doc is not None
        gu_a: str
        gu_x: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        gt_a: str
        gt_x: int
        gt_y: float
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # mapping

        def u_mapper(doc: dict[str, Any]) -> str:
            return f"{doc['p_bigint'] % 100}"

        def t_mapper(doc: MyTestDoc) -> str:
            return f"{doc['p_bigint'] % 100}"

        u_cur_1m = g_co_untyped.find().map(u_mapper)
        t_cur_1m = g_co_typed.find().map(t_mapper)
        gmu = next(u_cur_1m)
        gmt = next(t_cur_1m)
        assert gmu is not None
        assert gmt is not None
        gmu_a: str
        gmu_x: int
        gmu_a = gmu  # noqa: F841
        gmu_x = gmu  # type: ignore[assignment]  # noqa: F841
        gmt_a: str
        gmt_x: int
        gmt_a = gmu  # noqa: F841
        gmt_x = gmt  # type: ignore[assignment]  # noqa: F841

        # mapping composition

        def mapper_2(strint: str) -> float:
            return float(strint) / 10.0

        u_cur_2m = g_co_untyped.find().map(u_mapper).map(mapper_2)
        t_cur_2m = g_co_typed.find().map(t_mapper).map(mapper_2)
        gmu2 = next(u_cur_2m)
        gmt2 = next(t_cur_2m)
        assert gmu2 is not None
        assert gmt2 is not None
        gmu_a2: float
        gmu_x2: str
        gmu_a2 = gmu2  # noqa: F841
        gmu_x2 = gmu2  # type: ignore[assignment]  # noqa: F841
        gmt_a2: float
        gmt_x2: str
        gmt_a2 = gmu2  # noqa: F841
        gmt_x2 = gmt2  # type: ignore[assignment]  # noqa: F841

        # cloning a cursor
        u_cur_3c = g_co_untyped.find().map(u_mapper).clone()
        t_cur_3c = g_co_typed.find().map(t_mapper).clone()
        gu_doc3 = next(u_cur_3c)
        gt_doc3 = next(t_cur_3c)
        assert gu_doc3 is not None
        assert gt_doc3 is not None
        gu_a3: str
        gu_x3: int
        gu_a3 = gu_doc3  # noqa: F841
        gu_x3 = gu_doc3  # type: ignore[assignment] # noqa: F841
        gt_a3: str
        gt_x3: int
        gt_a3 = gt_doc3  # noqa: F841
        gt_x3 = gt_doc3  # type: ignore[assignment]  # noqa: F841

        # reading the buffer
        u_cur_4rb = g_co_untyped.find().map(u_mapper)
        t_cur_4rb = g_co_typed.find().map(t_mapper)
        next(u_cur_4rb)
        next(t_cur_4rb)
        u_doc_rbuf4 = u_cur_4rb.consume_buffer(3)[0]
        t_doc_rbuf4 = t_cur_4rb.consume_buffer(3)[0]
        assert u_doc_rbuf4 is not None
        assert t_doc_rbuf4 is not None
        gu_rbuf_a: str
        gu_rbuf_x: int
        gu_rbuf_a = u_doc_rbuf4["p_ascii"]  # noqa: F841
        gu_rbuf_x = u_doc_rbuf4["p_ascii"]  # noqa: F841
        gt_rbuf_a: str
        gt_rbuf_x: int
        gt_rbuf_y: float
        gt_rbuf_a = t_doc_rbuf4["p_ascii"]  # noqa: F841
        gt_rbuf_x = t_doc_rbuf4["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_rbuf_y = t_doc_rbuf4["c"]  # type: ignore[typeddict-item]  # noqa: F841

        g_co_typed.delete_many({})

    @pytest.mark.skipif(
        not RUN_RERANKER_TESTS,
        reason="Reranker-related tests skipped (no Astra and no header reranker key provided)",
    )
    @pytest.mark.describe("test of typing find_and_rerank, sync")
    def test_collection_find_and_rerank_typing_sync(
        self,
        sync_database: Database,
        sync_empty_collection: DefaultCollection,
        service_collection_parameters: dict[str, Any],
    ) -> None:
        """Test of typing in find_and_rerank, sync."""

        params = service_collection_parameters

        # Untyped baseline
        f_co_untyped = sync_database.get_collection(
            sync_empty_collection.name,
            reranking_api_key=params["reranking_api_key"]
            if USE_RERANKER_API_KEY_HEADER
            else _UNSET,
        )
        f_co_untyped.insert_one(VLEX_DOCUMENT)
        farr_u_hits = f_co_untyped.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_u_hits) > 0
        farr_u_doc = farr_u_hits[0].document
        fu_a: str
        fu_b: int
        fu_a = farr_u_doc["p_ascii"]  # noqa: F841
        fu_b = farr_u_doc["p_bigint"]  # noqa: F841
        # untyped, these are all ok:
        fu_x: int
        fu_y: float
        fu_x = farr_u_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            fu_y = farr_u_doc["c"]  # noqa: F841

        f_co_untyped.delete_many({})

        # Typed
        f_co_typed: Collection[MyTestDoc] = sync_database.get_collection(
            sync_empty_collection.name,
            document_type=MyTestDoc,
            reranking_api_key=params["reranking_api_key"]
            if USE_RERANKER_API_KEY_HEADER
            else _UNSET,
        )
        f_co_typed.insert_one(TYPED_VLEX_DOCUMENT)
        farr_t_hits = f_co_typed.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_t_hits) > 0
        farr_t_doc = farr_t_hits[0].document
        ft_a: str
        ft_b: int
        ft_a = farr_t_doc["p_ascii"]  # noqa: F841
        ft_b = farr_t_doc["p_bigint"]  # noqa: F841
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ft_x: int
        ft_y: float
        ft_x = farr_t_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ft_y = farr_t_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        farr_tm_hits = f_co_typed.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            projection=FIND_PROJECTION,
            document_type=MyTestMiniDoc,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_tm_hits) > 0
        farr_tm_doc = farr_tm_hits[0].document
        ftm_a: str
        ftm_b: int
        # typechecks must detect that p_ascii is not there
        with pytest.raises(KeyError):
            ftm_a = farr_tm_doc["p_ascii"]  # type: ignore[typeddict-item]  # noqa: F841
        ftm_b = farr_tm_doc["p_bigint"]  # noqa: F841

        f_co_typed.delete_many({})

    @pytest.mark.describe("test of typing create_collection, async")
    async def test_create_collection_typing_async(
        self,
        async_database: AsyncDatabase,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        """Test of creating typed collections with generics (and not), async."""

        # Untyped baseline
        ac_co_untyped = await async_database.create_collection(
            async_empty_collection.name,
            definition=(
                CollectionDefinition.builder()
                .set_vector_dimension(2)
                .set_vector_metric(VectorMetric.COSINE)
                .set_indexing("deny", ["not_indexed"])
                .build()
            ),
        )
        await ac_co_untyped.insert_one(DOCUMENT)
        cu_doc = await ac_co_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"_id", "p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_co_untyped_cursor = ac_co_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = await c_co_untyped_cursor.__anext__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        await ac_co_untyped.delete_many({})

        # Typed
        ac_co_typed: AsyncCollection[
            MyTestDoc
        ] = await async_database.create_collection(
            async_empty_collection.name,
            definition=(
                CollectionDefinition.builder()
                .set_vector_dimension(2)
                .set_vector_metric(VectorMetric.COSINE)
                .set_indexing("deny", ["not_indexed"])
                .build()
            ),
            document_type=MyTestDoc,
        )
        await ac_co_typed.insert_one(TYPED_DOCUMENT)
        ct_doc = await ac_co_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"_id", "p_ascii", "p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_co_typed_cursor = ac_co_typed.find(
            {}, projection=FIND_PROJECTION, document_type=MyTestMiniDoc
        )
        ctcur_doc = await c_co_typed_cursor.__anext__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        await ac_co_typed.delete_many({})

    @pytest.mark.describe("test of typing get_collection, async")
    async def test_get_collection_typing_async(
        self,
        async_database: AsyncDatabase,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        """Test of getting typed collections with generics (and not), async."""

        # Untyped baseline
        ag_co_untyped = async_database.get_collection(async_empty_collection.name)
        await ag_co_untyped.insert_one(DOCUMENT)
        gu_doc = await ag_co_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"_id", "p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        await ag_co_untyped.delete_many({})

        # Typed
        ag_co_typed: AsyncCollection[MyTestDoc] = async_database.get_collection(
            async_empty_collection.name,
            document_type=MyTestDoc,
        )
        await ag_co_typed.insert_one(TYPED_DOCUMENT)
        gt_doc = await ag_co_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"_id", "p_ascii", "p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        await ag_co_typed.delete_many({})

    @pytest.mark.describe("test of typing collection cursors with map, async")
    async def test_collection_cursormap_typing_async(
        self,
        async_database: AsyncDatabase,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        """Test of getting typed collections with generics (and not), sync."""

        ag_co_untyped = async_database.get_collection(async_empty_collection.name)
        ag_co_typed: AsyncCollection[MyTestDoc] = async_database.get_collection(
            async_empty_collection.name,
            document_type=MyTestDoc,
        )
        await ag_co_untyped.insert_many([DOCUMENT] * 30)

        # base cursors, read from them (un/typed)
        u_cur0 = ag_co_untyped.find()
        t_cur0 = ag_co_typed.find()
        gu_doc = await u_cur0.__anext__()
        gt_doc = await t_cur0.__anext__()
        assert gu_doc is not None
        assert gt_doc is not None
        gu_a: str
        gu_x: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        gt_a: str
        gt_x: int
        gt_y: float
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # mapping

        def u_mapper(doc: dict[str, Any]) -> str:
            return f"{doc['p_bigint'] % 100}"

        def t_mapper(doc: MyTestDoc) -> str:
            return f"{doc['p_bigint'] % 100}"

        u_cur_1m = ag_co_untyped.find().map(u_mapper)
        t_cur_1m = ag_co_typed.find().map(t_mapper)
        gmu = await u_cur_1m.__anext__()
        gmt = await t_cur_1m.__anext__()
        assert gmu is not None
        assert gmt is not None
        gmu_a: str
        gmu_x: int
        gmu_a = gmu  # noqa: F841
        gmu_x = gmu  # type: ignore[assignment]  # noqa: F841
        gmt_a: str
        gmt_x: int
        gmt_a = gmu  # noqa: F841
        gmt_x = gmt  # type: ignore[assignment]  # noqa: F841

        # mapping composition

        def mapper_2(strint: str) -> float:
            return float(strint) / 10.0

        u_cur_2m = ag_co_untyped.find().map(u_mapper).map(mapper_2)
        t_cur_2m = ag_co_typed.find().map(t_mapper).map(mapper_2)
        gmu2 = await u_cur_2m.__anext__()
        gmt2 = await t_cur_2m.__anext__()
        assert gmu2 is not None
        assert gmt2 is not None
        gmu_a2: float
        gmu_x2: str
        gmu_a2 = gmu2  # noqa: F841
        gmu_x2 = gmu2  # type: ignore[assignment]  # noqa: F841
        gmt_a2: float
        gmt_x2: str
        gmt_a2 = gmu2  # noqa: F841
        gmt_x2 = gmt2  # type: ignore[assignment]  # noqa: F841

        # cloning a cursor
        u_cur_3c = ag_co_untyped.find().map(u_mapper).clone()
        t_cur_3c = ag_co_typed.find().map(t_mapper).clone()
        gu_doc3 = await u_cur_3c.__anext__()
        gt_doc3 = await t_cur_3c.__anext__()
        assert gu_doc3 is not None
        assert gt_doc3 is not None
        gu_a3: str
        gu_x3: int
        gu_a3 = gu_doc3  # noqa: F841
        gu_x3 = gu_doc3  # type: ignore[assignment] # noqa: F841
        gt_a3: str
        gt_x3: int
        gt_a3 = gt_doc3  # noqa: F841
        gt_x3 = gt_doc3  # type: ignore[assignment]  # noqa: F841

        # reading the buffer
        u_cur_4rb = ag_co_untyped.find().map(u_mapper)
        t_cur_4rb = ag_co_typed.find().map(t_mapper)
        await u_cur_4rb.__anext__()
        await t_cur_4rb.__anext__()
        u_doc_rbuf4 = u_cur_4rb.consume_buffer(3)[0]
        t_doc_rbuf4 = t_cur_4rb.consume_buffer(3)[0]
        assert u_doc_rbuf4 is not None
        assert t_doc_rbuf4 is not None
        gu_rbuf_a: str
        gu_rbuf_x: int
        gu_rbuf_a = u_doc_rbuf4["p_ascii"]  # noqa: F841
        gu_rbuf_x = u_doc_rbuf4["p_ascii"]  # noqa: F841
        gt_rbuf_a: str
        gt_rbuf_x: int
        gt_rbuf_y: float
        gt_rbuf_a = t_doc_rbuf4["p_ascii"]  # noqa: F841
        gt_rbuf_x = t_doc_rbuf4["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_rbuf_y = t_doc_rbuf4["c"]  # type: ignore[typeddict-item]  # noqa: F841

        await ag_co_typed.delete_many({})

    @pytest.mark.skipif(
        not RUN_RERANKER_TESTS,
        reason="Reranker-related tests skipped (no Astra and no header reranker key provided)",
    )
    @pytest.mark.describe("test of typing find_and_rerank, async")
    async def test_collection_find_and_rerank_typing_async(
        self,
        async_database: AsyncDatabase,
        async_empty_collection: DefaultAsyncCollection,
        service_collection_parameters: dict[str, Any],
    ) -> None:
        """Test of typing in find_and_rerank, sync."""

        params = service_collection_parameters

        # Untyped baseline
        f_co_untyped = async_database.get_collection(
            async_empty_collection.name,
            reranking_api_key=params["reranking_api_key"]
            if USE_RERANKER_API_KEY_HEADER
            else _UNSET,
        )
        await f_co_untyped.insert_one(VLEX_DOCUMENT)
        farr_u_hits = await f_co_untyped.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_u_hits) > 0
        farr_u_doc = farr_u_hits[0].document
        fu_a: str
        fu_b: int
        fu_a = farr_u_doc["p_ascii"]  # noqa: F841
        fu_b = farr_u_doc["p_bigint"]  # noqa: F841
        # untyped, these are all ok:
        fu_x: int
        fu_y: float
        fu_x = farr_u_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            fu_y = farr_u_doc["c"]  # noqa: F841

        await f_co_untyped.delete_many({})

        # Typed
        af_co_typed: AsyncCollection[MyTestDoc] = async_database.get_collection(
            async_empty_collection.name,
            document_type=MyTestDoc,
            reranking_api_key=params["reranking_api_key"]
            if USE_RERANKER_API_KEY_HEADER
            else _UNSET,
        )
        await af_co_typed.insert_one(TYPED_VLEX_DOCUMENT)
        farr_t_hits = await af_co_typed.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_t_hits) > 0
        farr_t_doc = farr_t_hits[0].document
        ft_a: str
        ft_b: int
        ft_a = farr_t_doc["p_ascii"]  # noqa: F841
        ft_b = farr_t_doc["p_bigint"]  # noqa: F841
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ft_x: int
        ft_y: float
        ft_x = farr_t_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ft_y = farr_t_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        farr_tm_hits = await af_co_typed.find_and_rerank(
            sort={"$hybrid": {"$vector": [2, 1], "$lexical": "bla"}},
            limit=1,
            projection=FIND_PROJECTION,
            document_type=MyTestMiniDoc,
            rerank_on="p_ascii",
            rerank_query="bli",
        ).to_list()
        assert len(farr_tm_hits) > 0
        farr_tm_doc = farr_tm_hits[0].document
        ftm_a: str
        ftm_b: int
        # typechecks must detect that p_ascii is not there
        with pytest.raises(KeyError):
            ftm_a = farr_tm_doc["p_ascii"]  # type: ignore[typeddict-item]  # noqa: F841
        ftm_b = farr_tm_doc["p_bigint"]  # noqa: F841

        await af_co_typed.delete_many({})



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from astrapy.event_observers.context_managers import event_collector
from astrapy.event_observers.events import (
    ObservableError,
    ObservableEvent,
    ObservableEventType,
    ObservableRequest,
    ObservableResponse,
    ObservableWarning,
)
from astrapy.event_observers.observers import Observer

__all__ = (
    "event_collector",
    "ObservableEventType",
    "ObservableEvent",
    "ObservableError",
    "ObservableWarning",
    "ObservableRequest",
    "ObservableResponse",
    "Observer",
)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import logging
import re
import ssl
import weakref
from collections.abc import Iterable, Sequence
from decimal import Decimal
from types import TracebackType
from typing import Any, cast

import certifi
import httpx
from uuid6 import uuid7

from astrapy.constants import CallerType
from astrapy.event_observers import (
    ObservableError,
    ObservableRequest,
    ObservableResponse,
    ObservableWarning,
    Observer,
)
from astrapy.exceptions import (
    DataAPIHttpException,
    DataAPIResponseException,
    DevOpsAPIHttpException,
    DevOpsAPIResponseException,
    UnexpectedDataAPIResponseException,
    UnexpectedDevOpsAPIResponseException,
    _TimeoutContext,
    to_dataapi_timeout_exception,
    to_devopsapi_timeout_exception,
)
from astrapy.exceptions.error_descriptors import (
    DataAPIErrorDescriptor,
    DataAPIWarningDescriptor,
)
from astrapy.settings.defaults import (
    CHECK_DECIMAL_ESCAPING_CONSISTENCY,
    DEFAULT_REDACTED_HEADER_NAMES,
    FIXED_SECRET_PLACEHOLDER,
)
from astrapy.utils.meta import issue_plain_warning
from astrapy.utils.python_version import get_python_version
from astrapy.utils.request_tools import (
    HttpMethod,
    log_httpx_request,
    log_httpx_response,
    to_httpx_timeout,
)
from astrapy.utils.user_agents import (
    compose_full_user_agent,
    detect_astrapy_user_agent,
)

user_agent_astrapy = detect_astrapy_user_agent()

PYTHON_VERSION_SSL_ISSUES_WARNING = (
    "SSL connection reuse disabled due to a Python 3.12.[0-11] bug. "
    "This may reduce performance under certain workloads. "
    "Please upgrade to Python 3.12.12 or newer if possible."
)
CLIENT_SSL_CONTEXT = ssl.create_default_context(
    cafile=certifi.where()
)  # portable CA roots

logger = logging.getLogger(__name__)

# Used only for a range of Python versions where a SSL bug is detected.
no_pooling_limits = httpx.Limits(max_keepalive_connections=0, keepalive_expiry=0)
disable_ssl_reuse: bool
python_version = get_python_version()
if python_version >= (3, 12, 0) and python_version < (3, 12, 12):
    issue_plain_warning(
        PYTHON_VERSION_SSL_ISSUES_WARNING,
        stacklevel=3,
    )
    disable_ssl_reuse = True
else:
    disable_ssl_reuse = False

# these are a mixture from disparate alphabet, to minimize the chance
# of a collision with user-provided actual content:
DECIMAL_MARKER_PREFIX_STR = "𐐏丂"
DECIMAL_MARKER_SUFFIX_STR = "∀🇦🇫"
DECIMAL_CLEANER_PATTERN = re.compile(
    f'"{DECIMAL_MARKER_PREFIX_STR}([0-9.]+){DECIMAL_MARKER_SUFFIX_STR}"'
)


class _MarkedDecimalDefuser(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        if isinstance(obj, Decimal):
            return "(defused decimal)"
        return super().default(obj)


class _MarkedDecimalEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        if isinstance(obj, Decimal):
            return f"{DECIMAL_MARKER_PREFIX_STR}{obj}{DECIMAL_MARKER_SUFFIX_STR}"
        return super().default(obj)

    @staticmethod
    def _check_mark_match(json_string: str) -> bool:
        return bool(DECIMAL_CLEANER_PATTERN.search(json_string))

    @staticmethod
    def _clean_encoded_string(json_string: str) -> str:
        return re.sub(DECIMAL_CLEANER_PATTERN, r"\1", json_string)


class APICommander:
    client: httpx.Client
    async_client: httpx.AsyncClient

    def __init__(
        self,
        *,
        api_endpoint: str,
        path: str,
        spawner: object | None,
        headers: dict[str, str | None] = {},
        callers: Sequence[CallerType] = [],
        redacted_header_names: Iterable[str] | None = None,
        dev_ops_api: bool = False,
        event_observers: dict[str, Observer | None] = {},
        handle_decimals_writes: bool = False,
        handle_decimals_reads: bool = False,
    ) -> None:
        ssl_control_headers: dict[str, str | None]
        if disable_ssl_reuse:
            self.client = httpx.Client(
                limits=no_pooling_limits,
                verify=CLIENT_SSL_CONTEXT,
            )
            self.async_client = httpx.AsyncClient(
                limits=no_pooling_limits,
                verify=CLIENT_SSL_CONTEXT,
            )
            ssl_control_headers = {"Connection": "close"}
        else:
            self.client = httpx.Client(verify=CLIENT_SSL_CONTEXT)
            self.async_client = httpx.AsyncClient(verify=CLIENT_SSL_CONTEXT)
            ssl_control_headers = {}

        self.api_endpoint = api_endpoint.rstrip("/")
        self.path = path.lstrip("/")
        self.headers: dict[str, str | None] = {**ssl_control_headers, **headers}
        self.callers = callers
        self.redacted_header_names = set(redacted_header_names or [])
        self.upper_full_redacted_header_names = {
            header_name.upper()
            for header_name in (
                self.redacted_header_names | DEFAULT_REDACTED_HEADER_NAMES
            )
        }
        self.dev_ops_api = dev_ops_api
        self.event_observers = event_observers
        self.spawner_ref: weakref.ReferenceType[object] | None
        if spawner:
            self.spawner_ref = weakref.ref(spawner)
        else:
            self.spawner_ref = None
        self.handle_decimals_writes = handle_decimals_writes
        self.handle_decimals_reads = handle_decimals_reads

        self._faulty_response_exc_class: (
            type[UnexpectedDevOpsAPIResponseException]
            | type[UnexpectedDataAPIResponseException]
        )
        self._response_exc_class: (
            type[DevOpsAPIResponseException] | type[DataAPIResponseException]
        )
        self._http_exc_class: type[DataAPIHttpException] | type[DevOpsAPIHttpException]
        if self.dev_ops_api:
            self._faulty_response_exc_class = UnexpectedDevOpsAPIResponseException
            self._response_exc_class = DevOpsAPIResponseException
            self._http_exc_class = DevOpsAPIHttpException
        else:
            self._faulty_response_exc_class = UnexpectedDataAPIResponseException
            self._response_exc_class = DataAPIResponseException
            self._http_exc_class = DataAPIHttpException
        self._api_description = "DevOps API" if self.dev_ops_api else "Data API"

        full_user_agent_string = compose_full_user_agent(
            list(self.callers) + [user_agent_astrapy]
        )
        self.caller_header: dict[str, str] = (
            {"User-Agent": full_user_agent_string} if full_user_agent_string else {}
        )
        self.full_headers: dict[str, str] = {
            k: v
            for k, v in {
                **{
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
                **self.caller_header,
                **self.headers,
            }.items()
            if v is not None
        }
        self._loggable_headers = {
            k: v
            if k.upper() not in self.upper_full_redacted_header_names
            else FIXED_SECRET_PLACEHOLDER
            for k, v in self.full_headers.items()
        }
        self.full_path = ("/".join([self.api_endpoint, self.path])).rstrip("/")

    def __repr__(self) -> str:
        pieces = [
            f"api_endpoint={self.api_endpoint}",
            f"path={self.path}",
            f"callers={self.callers}",
            f"dev_ops_api={self.dev_ops_api}",
        ]
        inner_desc = ", ".join(pieces)
        return f"{self.__class__.__name__}({inner_desc})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, APICommander):
            return all(
                [
                    self.api_endpoint == other.api_endpoint,
                    self.path == other.path,
                    self.headers == other.headers,
                    self.callers == other.callers,
                    self.redacted_header_names == other.redacted_header_names,
                    self.dev_ops_api == other.dev_ops_api,
                ]
            )
        else:
            return False

    async def __aenter__(self) -> APICommander:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self.async_client.aclose()

    def _get_spawner(self) -> object | None:
        if self.spawner_ref is None:
            return None
        return self.spawner_ref()

    def _copy(
        self,
        api_endpoint: str | None = None,
        path: str | None = None,
        headers: dict[str, str | None] | None = None,
        callers: Sequence[CallerType] | None = None,
        redacted_header_names: list[str] | None = None,
        dev_ops_api: bool | None = None,
    ) -> APICommander:
        # some care in allowing e.g. {} to override (but not None):
        return APICommander(
            api_endpoint=(
                api_endpoint if api_endpoint is not None else self.api_endpoint
            ),
            path=path if path is not None else self.path,
            spawner=self._get_spawner(),
            headers=headers if headers is not None else self.headers,
            callers=callers if callers is not None else self.callers,
            redacted_header_names=(
                redacted_header_names
                if redacted_header_names is not None
                else self.redacted_header_names
            ),
            dev_ops_api=dev_ops_api if dev_ops_api is not None else self.dev_ops_api,
        )

    def _compose_request_url(self, additional_path: str | None) -> str:
        if additional_path:
            return "/".join([self.full_path.rstrip("/"), additional_path.lstrip("/")])
        else:
            return self.full_path

    def _raw_response_to_json(
        self,
        raw_response: httpx.Response,
        raise_api_errors: bool,
        payload: dict[str, Any] | None,
        caller_function_name: str | None,
        request_id: str,
    ) -> dict[str, Any]:
        # try to process the httpx raw response into a JSON or throw a failure
        raw_response_json: dict[str, Any]
        try:
            if self.handle_decimals_reads:
                # for decimal-aware contents (aka 'tables'), all number-looking things
                # are made into Decimal.
                # (for collections, this will be it. for Tables, schema-aware
                # proper post-processing will refine types, e.g. back to int, ...)
                raw_response_json = self._decimal_aware_parse_json_response(
                    raw_response.text,
                )
            else:
                raw_response_json = self._decimal_unaware_parse_json_response(
                    raw_response.text,
                )
        except ValueError:
            # json() parsing has failed (e.g., empty body)
            if payload is not None:
                command_desc = "/".join(sorted(payload.keys()))
            else:
                command_desc = "(none)"
            raise self._faulty_response_exc_class(
                text=f"Unparseable response from API '{command_desc}' command.",
                raw_response={
                    "raw_response": raw_response.text,
                },
            )

        # no warnings check for DevOps API (there, 'status' may contain a string)
        dictforced_response: dict[str, Any] = (
            raw_response_json if isinstance(raw_response_json, dict) else {}
        )
        if not self.dev_ops_api:
            warning_items: list[str | dict[str, Any]] = (
                dictforced_response.get("status") or {}
            ).get("warnings") or []
            if warning_items:
                warning_descriptors = [
                    DataAPIWarningDescriptor(warning_item)
                    for warning_item in warning_items
                ]
                if self.event_observers:
                    wrn_events = [
                        ObservableWarning(warning=warning_descriptor)
                        for warning_descriptor in warning_descriptors
                    ]
                    sender = self._get_spawner()
                    for ev_obs in self.event_observers.values():
                        if ev_obs is not None and ev_obs.enabled:
                            for wrn_event in wrn_events:
                                ev_obs.receive(
                                    wrn_event,
                                    sender=sender,
                                    function_name=caller_function_name,
                                    request_id=request_id,
                                )
                for warning_descriptor in warning_descriptors:
                    full_warning = (
                        f"The {self._api_description} returned "
                        f"a warning: {warning_descriptor}"
                    )
                    logger.warning(full_warning)

        if "errors" in dictforced_response:
            if self.event_observers:
                err_events = [
                    ObservableError(error=DataAPIErrorDescriptor(err_dict))
                    for err_dict in dictforced_response["errors"]
                ]
                sender = self._get_spawner()
                for ev_obs in self.event_observers.values():
                    if ev_obs is not None and ev_obs.enabled:
                        for err_event in err_events:
                            ev_obs.receive(
                                err_event,
                                sender=sender,
                                function_name=caller_function_name,
                                request_id=request_id,
                            )
            if raise_api_errors:
                logger.warning(
                    f"APICommander about to raise from: {dictforced_response['errors']}"
                )
                raise self._response_exc_class.from_response(
                    command=payload,
                    raw_response=dictforced_response,
                )

        return raw_response_json

    @staticmethod
    def _decimal_unaware_parse_json_response(response_text: str) -> dict[str, Any]:
        return cast(
            dict[str, Any],
            json.loads(response_text),
        )

    @staticmethod
    def _decimal_aware_parse_json_response(response_text: str) -> dict[str, Any]:
        return cast(
            dict[str, Any],
            json.loads(
                response_text,
                parse_float=Decimal,
                parse_int=Decimal,
            ),
        )

    @staticmethod
    def _decimal_unaware_encode_payload(payload: dict[str, Any] | None) -> str | None:
        # This is the JSON encoder in absence of the workaround to treat Decimals
        if payload is not None:
            return json.dumps(
                payload,
                allow_nan=False,
                separators=(",", ":"),
                ensure_ascii=False,
            )
        else:
            return None

    @staticmethod
    def _decimal_aware_encode_payload(payload: dict[str, Any] | None) -> str | None:
        if payload is not None:
            if CHECK_DECIMAL_ESCAPING_CONSISTENCY:
                # check if escaping collision. This is expensive and 99.9999999% useless
                _naive_dump = json.dumps(
                    payload,
                    allow_nan=False,
                    separators=(",", ":"),
                    ensure_ascii=False,
                    cls=_MarkedDecimalDefuser,
                )
                if _MarkedDecimalEncoder._check_mark_match(_naive_dump):
                    raise ValueError(
                        "The pattern to work around Decimals was detected in a "
                        "user-provided item. This payload cannot be JSON-encoded."
                    )
            dec_marked_dump = json.dumps(
                payload,
                allow_nan=False,
                separators=(",", ":"),
                ensure_ascii=False,
                cls=_MarkedDecimalEncoder,
            )
            return _MarkedDecimalEncoder._clean_encoded_string(dec_marked_dump)
        else:
            return None

    def raw_request(
        self,
        *,
        caller_function_name: str | None,
        request_id: str | None,
        http_method: str = HttpMethod.POST,
        payload: dict[str, Any] | None = None,
        additional_path: str | None = None,
        request_params: dict[str, Any] = {},
        timeout_context: _TimeoutContext | None = None,
    ) -> httpx.Response:
        if request_id is None:
            request_id = str(uuid7())
        request_url = self._compose_request_url(additional_path)
        _timeout_context = timeout_context or _TimeoutContext(request_ms=None)
        encoded_payload = (
            self._decimal_aware_encode_payload(payload)
            if self.handle_decimals_writes
            else self._decimal_unaware_encode_payload(payload)
        )
        log_httpx_request(
            http_method=http_method,
            full_url=request_url,
            request_params=request_params,
            redacted_request_headers=self._loggable_headers,
            encoded_payload=encoded_payload,
            timeout_context=_timeout_context,
            caller_function_name=caller_function_name,
        )
        httpx_timeout_s = to_httpx_timeout(_timeout_context)
        if self.event_observers:
            req_event = ObservableRequest(
                payload=encoded_payload,
                http_method=http_method,
                url=request_url,
                query_parameters=request_params,
                redacted_headers=self._loggable_headers,
                dev_ops_api=self.dev_ops_api,
            )
            sender = self._get_spawner()
            for ev_obs in self.event_observers.values():
                if ev_obs is not None and ev_obs.enabled:
                    ev_obs.receive(
                        req_event,
                        sender=sender,
                        function_name=caller_function_name,
                        request_id=request_id,
                    )

        try:
            raw_response = self.client.request(
                method=http_method,
                url=request_url,
                content=encoded_payload.encode()
                if encoded_payload is not None
                else None,
                params=request_params,
                timeout=httpx_timeout_s,
                headers=self.full_headers,
            )
        except httpx.TimeoutException as timeout_exc:
            if self.dev_ops_api:
                raise to_devopsapi_timeout_exception(
                    timeout_exc, timeout_context=_timeout_context
                )
            else:
                raise to_dataapi_timeout_exception(
                    timeout_exc, timeout_context=_timeout_context
                )

        if self.event_observers:
            rsp_event = ObservableResponse(
                body=raw_response.text,
                status_code=raw_response.status_code,
            )
            for ev_obs in self.event_observers.values():
                if ev_obs is not None and ev_obs.enabled:
                    ev_obs.receive(
                        rsp_event,
                        sender=self._get_spawner(),
                        function_name=caller_function_name,
                        request_id=request_id,
                    )
        try:
            raw_response.raise_for_status()
        except httpx.HTTPStatusError as http_exc:
            raise self._http_exc_class.from_httpx_error(http_exc)
        log_httpx_response(response=raw_response)
        return raw_response

    async def async_raw_request(
        self,
        *,
        caller_function_name: str | None,
        request_id: str | None,
        http_method: str = HttpMethod.POST,
        payload: dict[str, Any] | None = None,
        additional_path: str | None = None,
        request_params: dict[str, Any] = {},
        timeout_context: _TimeoutContext | None = None,
    ) -> httpx.Response:
        if request_id is None:
            request_id = str(uuid7())
        request_url = self._compose_request_url(additional_path)
        _timeout_context = timeout_context or _TimeoutContext(request_ms=None)
        encoded_payload = (
            self._decimal_aware_encode_payload(payload)
            if self.handle_decimals_writes
            else self._decimal_unaware_encode_payload(payload)
        )
        log_httpx_request(
            http_method=http_method,
            full_url=request_url,
            request_params=request_params,
            redacted_request_headers=self._loggable_headers,
            encoded_payload=encoded_payload,
            timeout_context=_timeout_context,
            caller_function_name=caller_function_name,
        )
        httpx_timeout_s = to_httpx_timeout(_timeout_context)
        if self.event_observers:
            req_event = ObservableRequest(
                payload=encoded_payload,
                http_method=http_method,
                url=request_url,
                query_parameters=request_params,
                redacted_headers=self._loggable_headers,
                dev_ops_api=self.dev_ops_api,
            )
            sender = self._get_spawner()
            for ev_obs in self.event_observers.values():
                if ev_obs is not None and ev_obs.enabled:
                    ev_obs.receive(
                        req_event,
                        sender=sender,
                        function_name=caller_function_name,
                        request_id=request_id,
                    )

        try:
            raw_response = await self.async_client.request(
                method=http_method,
                url=request_url,
                content=encoded_payload.encode()
                if encoded_payload is not None
                else None,
                params=request_params,
                timeout=httpx_timeout_s,
                headers=self.full_headers,
            )
        except httpx.TimeoutException as timeout_exc:
            if self.dev_ops_api:
                raise to_devopsapi_timeout_exception(
                    timeout_exc, timeout_context=_timeout_context
                )
            else:
                raise to_dataapi_timeout_exception(
                    timeout_exc, timeout_context=_timeout_context
                )

        if self.event_observers:
            rsp_event = ObservableResponse(
                body=raw_response.text,
                status_code=raw_response.status_code,
            )
            sender = self._get_spawner()
            for ev_obs in self.event_observers.values():
                if ev_obs is not None and ev_obs.enabled:
                    ev_obs.receive(
                        rsp_event,
                        sender=sender,
                        function_name=caller_function_name,
                        request_id=request_id,
                    )
        try:
            raw_response.raise_for_status()
        except httpx.HTTPStatusError as http_exc:
            raise self._http_exc_class.from_httpx_error(http_exc)
        log_httpx_response(response=raw_response)
        return raw_response

    def request(
        self,
        *,
        http_method: str = HttpMethod.POST,
        payload: dict[str, Any] | None = None,
        additional_path: str | None = None,
        request_params: dict[str, Any] = {},
        raise_api_errors: bool = True,
        timeout_context: _TimeoutContext | None = None,
        caller_function_name: str | None = None,
    ) -> dict[str, Any]:
        request_id = str(uuid7())
        raw_response = self.raw_request(
            http_method=http_method,
            payload=payload,
            additional_path=additional_path,
            request_params=request_params,
            timeout_context=timeout_context,
            caller_function_name=caller_function_name,
            request_id=request_id,
        )
        return self._raw_response_to_json(
            raw_response,
            raise_api_errors=raise_api_errors,
            payload=payload,
            caller_function_name=caller_function_name,
            request_id=request_id,
        )

    async def async_request(
        self,
        *,
        http_method: str = HttpMethod.POST,
        payload: dict[str, Any] | None = None,
        additional_path: str | None = None,
        request_params: dict[str, Any] = {},
        raise_api_errors: bool = True,
        timeout_context: _TimeoutContext | None = None,
        caller_function_name: str | None = None,
    ) -> dict[str, Any]:
        request_id = str(uuid7())
        raw_response = await self.async_raw_request(
            http_method=http_method,
            payload=payload,
            additional_path=additional_path,
            request_params=request_params,
            timeout_context=timeout_context,
            caller_function_name=caller_function_name,
            request_id=request_id,
        )
        return self._raw_response_to_json(
            raw_response,
            raise_api_errors=raise_api_errors,
            payload=payload,
            caller_function_name=caller_function_name,
            request_id=request_id,
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.authentication import (
    RerankingAPIKeyHeaderProvider,
    coerce_reranking_headers_provider,
)
from astrapy.settings.defaults import RERANKING_HEADER_API_KEY


class TestRerankingHeadersProvider:
    @pytest.mark.describe("test of headers from RerankingAPIKeyHeaderProvider")
    def test_rerankingheadersprovider_static(self) -> None:
        ehp = RerankingAPIKeyHeaderProvider("x")
        assert {k.lower(): v for k, v in ehp.get_headers().items()} == {
            RERANKING_HEADER_API_KEY.lower(): "x"
        }

    @pytest.mark.describe("test of headers from empty RerankingAPIKeyHeaderProvider")
    def test_rerankingheadersprovider_null(self) -> None:
        ehp = RerankingAPIKeyHeaderProvider(None)
        assert ehp.get_headers() == {}

    @pytest.mark.describe("test of reranking headers provider coercion")
    def test_rerankingheadersprovider_coercion(self) -> None:
        """This doubles as equality test."""
        ehp_s = RerankingAPIKeyHeaderProvider("x")
        ehp_n = RerankingAPIKeyHeaderProvider(None)
        assert coerce_reranking_headers_provider(ehp_s) == ehp_s
        assert coerce_reranking_headers_provider(ehp_n) == ehp_n

        assert coerce_reranking_headers_provider("x") == ehp_s



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from astrapy.settings.definitions.definitions_admin import (
    DatabaseStatus,
    Environment,
    ModelStatus,
)
from astrapy.settings.definitions.definitions_data import (
    DefaultIdType,
    MapEncodingMode,
    ReturnDocument,
    SortMode,
    VectorMetric,
    normalize_optional_projection,
)
from astrapy.settings.definitions.definitions_types import (
    DOC,
    DOC2,
    ROW,
    ROW2,
    CallerType,
    DefaultDocumentType,
    DefaultRowType,
    FilterType,
    HybridSortType,
    ProjectionType,
    SerializerFunctionType,
    SortType,
    UDTDeserializerFunctionType,
)

__all__ = [
    "DatabaseStatus",
    "DefaultIdType",
    "Environment",
    "MapEncodingMode",
    "ModelStatus",
    "ReturnDocument",
    "SortMode",
    "VectorMetric",
    "DefaultDocumentType",
    "DefaultRowType",
    "ProjectionType",
    "SortType",
    "HybridSortType",
    "FilterType",
    "CallerType",
    "SerializerFunctionType",
    "UDTDeserializerFunctionType",
    "ROW",
    "ROW2",
    "DOC",
    "DOC2",
    "normalize_optional_projection",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class DataAPIErrorDescriptor:
    """
    An object representing a single error, as returned from the Data API,
    typically with an error code, a text message and other properties.

    This object is used to describe errors received from the Data API,
    in the form of HTTP-200 ("success") responses containing errors

    Depending on the API command semantics, responses may express partial
    successes with some errors (for instance, an insertMany command inserting
    most of the documents/rows, but failing on a couple of incompatible inputs).

    Attributes:
        error_code: a string code as found in the API error's "errorCode" field.
        message: the text found in the API error's "message" field.
        title:  the text found in the API error's "title" field.
        family:  the text found in the API error's "family" field.
        scope:  the text found in the API error's "scope" field.
        id:  the text found in the API error's "id" field.
        attributes: a dict with any further key-value pairs returned by the API.
    """

    title: str | None
    error_code: str | None
    message: str | None
    family: str | None
    scope: str | None
    id: str | None
    attributes: dict[str, Any]

    _known_dict_fields = {
        "title",
        "errorCode",
        "message",
        "family",
        "scope",
        "id",
    }

    def __init__(self, error_dict: dict[str, str] | str) -> None:
        if isinstance(error_dict, str):
            self.message = error_dict
            self.title = None
            self.error_code = None
            self.family = None
            self.scope = None
            self.id = None
            self.attributes = {}
        else:
            self.title = error_dict.get("title")
            self.error_code = error_dict.get("errorCode")
            self.message = error_dict.get("message")
            self.family = error_dict.get("family")
            self.scope = error_dict.get("scope")
            self.id = error_dict.get("id")
            self.attributes = {
                k: v for k, v in error_dict.items() if k not in self._known_dict_fields
            }

    def __repr__(self) -> str:
        pieces = [
            f"{self.title.__repr__()}" if self.title else None,
            f"error_code={self.error_code.__repr__()}" if self.error_code else None,
            f"message={self.message.__repr__()}" if self.message else None,
            f"family={self.family.__repr__()}" if self.family else None,
            f"scope={self.scope.__repr__()}" if self.scope else None,
            f"id={self.id.__repr__()}" if self.id else None,
            f"attributes={self.attributes.__repr__()}" if self.attributes else None,
        ]
        return f"{self.__class__.__name__}({', '.join(pc for pc in pieces if pc)})"

    def __str__(self) -> str:
        return self.summary()

    def summary(self) -> str:
        """
        Determine a string succinct description of this descriptor.

        The precise format of this summary is determined by which fields are set.
        """
        non_code_part: str | None
        if self.title:
            if self.message:
                non_code_part = f"{self.title}: {self.message}"
            else:
                non_code_part = f"{self.title}"
        else:
            if self.message:
                non_code_part = f"{self.message}"
            else:
                non_code_part = None
        if self.error_code:
            if non_code_part:
                return f"{non_code_part} ({self.error_code})"
            else:
                return f"{self.error_code}"
        else:
            if non_code_part:
                return non_code_part
            else:
                return ""


@dataclass
class DataAPIWarningDescriptor(DataAPIErrorDescriptor):
    """
    An object representing a single warning, as returned from the Data API,
    typically with a code, a text message and other properties.

    This object is used to describe warnings received from the Data API,
    in the form of HTTP-200 ("success") responses with accompanying warnings.

    Attributes:
        error_code: a string code found in the API warning's "errorCode" field.
        message: the text found in the API warning's "message" field.
        title:  the text found in the API warning's "title" field.
        family:  the text found in the API warning's "family" field.
        scope:  the text found in the API warning's "scope" field.
        id:  the text found in the API warning's "id" field.
        attributes: a dict with any further key-value pairs returned by the API.
    """

    def __init__(self, error_dict: dict[str, str] | str) -> None:
        return DataAPIErrorDescriptor.__init__(self, error_dict=error_dict)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import sys

from astrapy.info import EmbeddingProviderParameter, FindEmbeddingProvidersResult

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from live_provider_info import live_provider_info
from vectorize_models import live_test_models


def desc_param(param_data: EmbeddingProviderParameter) -> str:
    if param_data.parameter_type.lower() == "string":
        return "str"
    elif param_data.parameter_type.lower() == "number":
        validation = param_data.validation
        if "numericRange" in validation:
            validation_nr = validation["numericRange"]
            assert isinstance(validation_nr, list) and len(validation_nr) == 2
            range_desc = f"[{validation_nr[0]} : {validation_nr[1]}]"
            if param_data.default_value is not None:
                range_desc2 = f"{range_desc} (default={param_data.default_value})"
            else:
                range_desc2 = range_desc
            return f"number, {range_desc2}"
        elif "options" in validation:
            validation_op = validation["options"]
            assert isinstance(validation_op, list) and len(validation_op) > 1
            return f"number, {' / '.join(str(v) for v in validation_op)}"
        else:
            raise ValueError(
                f"Unknown number validation spec: '{json.dumps(validation)}'"
            )
    elif param_data.parameter_type.lower() == "boolean":
        return "bool"
    else:
        raise NotImplementedError


if __name__ == "__main__":
    provider_info: FindEmbeddingProvidersResult = live_provider_info()
    providers_json = (provider_info.raw_info or {}).get("embeddingProviders")
    if not providers_json:
        raise ValueError(
            "raw info from embedding providers lacks `embeddingProviders` content."
        )
    json.dump(providers_json, open("_providers.json", "w"), indent=2, sort_keys=True)

    for provider, provider_data in sorted(provider_info.embedding_providers.items()):
        print(f"{provider} ({len(provider_data.models)} models)")
        print("    auth:")
        for auth_type, auth_data in sorted(
            provider_data.supported_authentication.items()
        ):
            if auth_data.enabled:
                tokens = ", ".join(f"'{tok.accepted}'" for tok in auth_data.tokens)
                print(f"      {auth_type} ({tokens})")
        if provider_data.parameters:
            print("    parameters")
            for param_data in provider_data.parameters:
                param_name = param_data.name
                if param_data.required:
                    param_display_name = param_name
                else:
                    param_display_name = f"({param_name})"
                param_desc = desc_param(param_data)
                print(f"      - {param_display_name}: {param_desc}")
        print("    models:")
        for model_data in sorted(provider_data.models, key=lambda pro: pro.name):
            model_name = model_data.name
            if model_data.vector_dimension is not None:
                assert model_data.vector_dimension > 0
                model_dim_desc = f" (D = {model_data.vector_dimension})"
            else:
                model_dim_desc = ""
            if True:
                print(f"      {model_name}{model_dim_desc}")
                if model_data.parameters:
                    for param_data in model_data.parameters:
                        param_name = param_data.name
                        if param_data.required:
                            param_display_name = param_name
                        else:
                            param_display_name = f"({param_name})"
                        param_desc = desc_param(param_data)
                        print(f"        - {param_display_name}: {param_desc}")

    print("\n" * 2)
    all_test_models = list(live_test_models())
    for auth_type in ["HEADER", "NONE", "SHARED_SECRET"]:
        print(f"Tags for auth type {auth_type}:", end="")
        #
        at_test_models = [
            test_model
            for test_model in all_test_models
            if test_model["auth_type_name"] == auth_type
        ]
        at_model_ids: list[str] = sorted(
            [str(model_desc["model_tag"]) for model_desc in at_test_models]
        )
        if at_model_ids:
            print("")
            print("\n".join(f"    {ami}" for ami in at_model_ids))
        else:
            print(" (no tags)")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import time
from collections.abc import Awaitable
from typing import Any, Callable

import pytest

from astrapy import AsyncDatabase, DataAPIClient, Database
from astrapy.api_options import APIOptions, TimeoutOptions
from astrapy.settings.defaults import API_ENDPOINT_TEMPLATE_ENV_MAP

from ..conftest import (
    ADMIN_ENV_LIST,
    ADMIN_ENV_VARIABLE_MAP,
    IS_ASTRA_DB,
)

KEYSPACE_POLL_SLEEP_TIME = 2
KEYSPACE_TIMEOUT = 30
DATABASE_POLL_SLEEP_TIME = 10
DATABASE_TIMEOUT = 480
PRE_DROP_SAFETY_POLL_INTERVAL = 5
PRE_DROP_SAFETY_TIMEOUT = 120

TESTING_SAFE_CLIENT_TIMEOUT_MS = 600000 * 3
TESTING_SAFE_CLIENT_OPTIONS = APIOptions(
    timeout_options=TimeoutOptions(
        database_admin_timeout_ms=TESTING_SAFE_CLIENT_TIMEOUT_MS
    )
)


def admin_test_envs_tokens() -> list[Any]:
    """
    This actually returns a list of `_pytest.mark.structures.ParameterSet` instances,
    each wrapping a tuple[str, Optional[str]] = (env, token)
    """
    envs_tokens: list[Any] = []
    for admin_env in ADMIN_ENV_LIST:
        markers = []
        pair: tuple[str, str | None]
        if ADMIN_ENV_VARIABLE_MAP[admin_env]["token"]:
            pair = (admin_env, ADMIN_ENV_VARIABLE_MAP[admin_env]["token"])
        else:
            pair = (admin_env, None)
            markers.append(pytest.mark.skip(reason=f"{admin_env} token not available"))
        envs_tokens.append(pytest.param(pair, marks=markers))

    return envs_tokens


def wait_until_true(
    poll_interval: int, max_seconds: int, condition: Callable[..., bool]
) -> None:
    ini_time = time.time()
    while not condition():
        if time.time() - ini_time > max_seconds:
            raise ValueError("Timed out on condition.")
        time.sleep(poll_interval)


async def await_until_true(
    poll_interval: int, max_seconds: int, acondition: Callable[..., Awaitable[bool]]
) -> None:
    ini_time = time.time()
    while not (await acondition()):
        if time.time() - ini_time > max_seconds:
            raise ValueError("Timed out on condition.")
        await asyncio.sleep(poll_interval)


@pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
class TestAdmin:
    @pytest.mark.parametrize(
        "admin_env_token", admin_test_envs_tokens(), ids=ADMIN_ENV_LIST
    )
    @pytest.mark.describe("test of the full tour with AstraDBDatabaseAdmin, sync")
    def test_astra_db_database_admin_sync(
        self, admin_env_token: tuple[str, str]
    ) -> None:
        """
        Test plan (it has to be a single giant test to use one DB throughout):
        - create client -> get_admin
        - create a db (wait)
        - with the AstraDBDatabaseAdmin:
            - info
            - list keyspaces, check
            - create 2 keyspaces (wait, nonwait)
            - list keyspaces, check
            - get_database -> create_collection/list_collection_names
            - get_async_database, check if == previous
            - drop keyspaces (wait, nonwait)
            - list keyspaces, check
            - drop database (wait)
        - check DB not existing
        """
        admin_env, token = admin_env_token
        db_name = f"test_database_{admin_env}"
        db_provider = ADMIN_ENV_VARIABLE_MAP[admin_env]["provider"]
        db_region = ADMIN_ENV_VARIABLE_MAP[admin_env]["region"]

        if db_provider is None:
            raise ValueError("Cannot proceed with provider unset.")
        if db_region is None:
            raise ValueError("Cannot proceed with region unset.")

        # create client, get admin
        client: DataAPIClient
        if admin_env == "prod":
            client = DataAPIClient(token, api_options=TESTING_SAFE_CLIENT_OPTIONS)
        else:
            client = DataAPIClient(
                token, environment=admin_env, api_options=TESTING_SAFE_CLIENT_OPTIONS
            )
        admin = client.get_admin()

        # create a db (wait)
        db_admin = admin.create_database(
            name=db_name,
            keyspace="custom_keyspace",
            wait_until_active=True,
            cloud_provider=db_provider,
            region=db_region,
        )

        # info with the AstraDBDatabaseAdmin
        created_db_id = db_admin.id
        assert db_admin.info().id == created_db_id

        # list nss
        keyspaces1 = set(db_admin.list_keyspaces())
        assert keyspaces1 == {"custom_keyspace"}

        # create two keyspaces
        db_admin.create_keyspace(
            "waited_ks",
            wait_until_active=True,
        )

        db_admin.create_keyspace(
            "nonwaited_ks",
            wait_until_active=False,
        )
        wait_until_true(
            poll_interval=KEYSPACE_POLL_SLEEP_TIME,
            max_seconds=KEYSPACE_TIMEOUT,
            condition=lambda: "nonwaited_ks" in db_admin.list_keyspaces(),
        )

        keyspaces3 = set(db_admin.list_keyspaces())
        assert keyspaces3 - keyspaces1 == {"waited_ks", "nonwaited_ks"}

        # get db and use it
        db = db_admin.get_database(keyspace="custom_keyspace")
        db.create_collection("canary_coll")
        assert "canary_coll" in db.list_collection_names()

        # check async db is the same
        assert db_admin.get_async_database(keyspace="custom_keyspace").to_sync() == db

        # drop nss, wait, nonwait
        db_admin.drop_keyspace(
            "waited_ks",
            wait_until_active=True,
        )

        db_admin.drop_keyspace(
            "nonwaited_ks",
            wait_until_active=False,
        )
        wait_until_true(
            poll_interval=KEYSPACE_POLL_SLEEP_TIME,
            max_seconds=KEYSPACE_TIMEOUT,
            condition=lambda: "nonwaited_ks" not in db_admin.list_keyspaces(),
        )

        # check nss after dropping two of them
        keyspaces1b = set(db_admin.list_keyspaces())
        assert keyspaces1b == keyspaces1

        # drop db and check. We wait a little due to "nontransactional cluster md"
        wait_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            condition=lambda: db_admin.info().status == "ACTIVE",
        )
        db_admin.drop()

        db_ids = {db.id for db in admin.list_databases()}
        assert created_db_id not in db_ids

    @pytest.mark.parametrize(
        "admin_env_token", admin_test_envs_tokens(), ids=ADMIN_ENV_LIST
    )
    @pytest.mark.describe(
        "test of the full tour with AstraDBAdmin and client methods, sync"
    )
    def test_astra_db_admin_sync(self, admin_env_token: tuple[str, str]) -> None:
        """
        Test plan (it has to be a single giant test to use the two DBs throughout):
        - create client -> get_admin
        - create two dbs (wait, nonwait)
        - list the two database ids, check
        - get check info on one such db through admin
        - with the client:
            4 get_dbs (a/sync, by id+region/api_endpoint), check if equal
            and test their list_collections
        - get_db_admin from the admin for one of the dbs
            - create ns
            - get_database -> list_collection_names
            - get_async_database and check == with above
        - drop dbs, (wait, nonwait)
        """
        admin_env, token = admin_env_token
        db_name_w = f"test_database_w_{admin_env}"
        db_name_nw = f"test_database_nw_{admin_env}"
        db_provider = ADMIN_ENV_VARIABLE_MAP[admin_env]["provider"]
        db_region = ADMIN_ENV_VARIABLE_MAP[admin_env]["region"]

        if db_provider is None:
            raise ValueError("Cannot proceed with provider unset.")
        if db_region is None:
            raise ValueError("Cannot proceed with region unset.")

        # create client and get admin
        client: DataAPIClient
        if admin_env == "prod":
            client = DataAPIClient(token, api_options=TESTING_SAFE_CLIENT_OPTIONS)
        else:
            client = DataAPIClient(
                token, environment=admin_env, api_options=TESTING_SAFE_CLIENT_OPTIONS
            )
        admin = client.get_admin()

        # create the two dbs
        db_admin_nw = admin.create_database(
            name=db_name_nw,
            wait_until_active=False,
            cloud_provider=db_provider,
            region=db_region,
        )
        created_db_id_nw = db_admin_nw.id
        db_admin_w = admin.create_database(
            name=db_name_w,
            wait_until_active=True,
            cloud_provider=db_provider,
            region=db_region,
        )
        created_db_id_w = db_admin_w.id

        def _waiter1() -> bool:
            db_ids = {db.id for db in admin.list_databases()}
            return created_db_id_nw in db_ids

        wait_until_true(
            poll_interval=DATABASE_POLL_SLEEP_TIME,
            max_seconds=DATABASE_TIMEOUT,
            condition=_waiter1,
        )

        # list, check ids
        db_ids = {db.id for db in admin.list_databases()}
        assert {created_db_id_nw, created_db_id_w} - db_ids == set()

        # get info through admin
        db_w_info = admin.database_info(created_db_id_w)
        assert db_w_info.id == created_db_id_w

        # get and compare dbs obtained by the client
        synthetic_api_endpoint = API_ENDPOINT_TEMPLATE_ENV_MAP[admin_env].format(
            database_id=created_db_id_w,
            region=db_region,
        )
        db_w_e = client.get_database(synthetic_api_endpoint)
        adb_w_e = client.get_async_database(synthetic_api_endpoint)
        assert isinstance(db_w_e.list_collection_names(), list)
        assert adb_w_e.to_sync() == db_w_e

        # get db admin from the admin and use it
        db_w_admin = admin.get_database_admin(created_db_id_w, region=db_region)
        db_w_admin.create_keyspace("additional_keyspace")
        db_w_from_admin = db_w_admin.get_database()
        assert isinstance(db_w_from_admin.list_collection_names(), list)
        adb_w_from_admin = db_w_admin.get_async_database()
        assert adb_w_from_admin.to_sync() == db_w_from_admin

        # drop databases: the w one through the admin, the nw using its db-admin
        #   (this covers most cases if combined with the
        #   (w, using db-admin) of test_astra_db_database_admin)
        assert db_admin_nw == admin.get_database_admin(
            created_db_id_nw,
            region=db_region,
        )
        # drop db and check. We wait a little due to "nontransactional cluster md"
        wait_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            condition=lambda: db_admin_nw.info().status == "ACTIVE",
        )
        wait_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            condition=lambda: db_admin_w.info().status == "ACTIVE",
        )
        db_admin_nw.drop(wait_until_active=False)
        admin.drop_database(created_db_id_w)

        def _waiter2() -> bool:
            db_ids = {db.id for db in admin.list_databases()}
            return created_db_id_nw not in db_ids

        wait_until_true(
            poll_interval=DATABASE_POLL_SLEEP_TIME,
            max_seconds=DATABASE_TIMEOUT,
            condition=_waiter2,
        )

    @pytest.mark.parametrize(
        "admin_env_token", admin_test_envs_tokens(), ids=ADMIN_ENV_LIST
    )
    @pytest.mark.describe("test of the full tour with AstraDBDatabaseAdmin, async")
    async def test_astra_db_database_admin_async(
        self, admin_env_token: tuple[str, str]
    ) -> None:
        """
        Test plan (it has to be a single giant test to use one DB throughout):
        - create client -> get_admin
        - create a db (wait)
        - with the AstraDBDatabaseAdmin:
            - info
            - list keyspaces, check
            - create 2 keyspaces (wait, nonwait)
            - list keyspaces, check
            - get_database -> create_collection/list_collection_names
            - get_async_database, check if == previous
            - drop keyspaces (wait, nonwait)
            - list keyspaces, check
            - drop database (wait)
        - check DB not existing
        """
        admin_env, token = admin_env_token
        db_name = f"test_database_{admin_env}"
        db_provider = ADMIN_ENV_VARIABLE_MAP[admin_env]["provider"]
        db_region = ADMIN_ENV_VARIABLE_MAP[admin_env]["region"]

        if db_provider is None:
            raise ValueError("Cannot proceed with provider unset.")
        if db_region is None:
            raise ValueError("Cannot proceed with region unset.")

        # create client, get admin
        client: DataAPIClient
        if admin_env == "prod":
            client = DataAPIClient(token, api_options=TESTING_SAFE_CLIENT_OPTIONS)
        else:
            client = DataAPIClient(
                token, environment=admin_env, api_options=TESTING_SAFE_CLIENT_OPTIONS
            )
        admin = client.get_admin()

        # create a db (wait)
        db_admin = await admin.async_create_database(
            name=db_name,
            keyspace="custom_keyspace",
            wait_until_active=True,
            cloud_provider=db_provider,
            region=db_region,
        )

        # info with the AstraDBDatabaseAdmin
        created_db_id = db_admin.id
        assert (await db_admin.async_info()).id == created_db_id

        # list nss
        keyspaces1 = set(await db_admin.async_list_keyspaces())
        assert keyspaces1 == {"custom_keyspace"}

        # create two keyspaces
        await db_admin.async_create_keyspace(
            "waited_ks",
            wait_until_active=True,
        )

        await db_admin.async_create_keyspace(
            "nonwaited_ks",
            wait_until_active=False,
        )

        async def _awaiter1() -> bool:
            return "nonwaited_ks" in (await db_admin.async_list_keyspaces())

        await await_until_true(
            poll_interval=KEYSPACE_POLL_SLEEP_TIME,
            max_seconds=KEYSPACE_TIMEOUT,
            acondition=_awaiter1,
        )

        keyspaces3 = set(await db_admin.async_list_keyspaces())
        assert keyspaces3 - keyspaces1 == {"waited_ks", "nonwaited_ks"}

        # get db and use it
        adb = db_admin.get_async_database(keyspace="custom_keyspace")
        await adb.create_collection("canary_coll")
        assert "canary_coll" in (await adb.list_collection_names())

        # check sync db is the same
        assert db_admin.get_database(keyspace="custom_keyspace").to_async() == adb

        # drop nss, wait, nonwait
        await db_admin.async_drop_keyspace(
            "waited_ks",
            wait_until_active=True,
        )

        await db_admin.async_drop_keyspace(
            "nonwaited_ks",
            wait_until_active=False,
        )

        async def _awaiter2() -> bool:
            ks_list = await db_admin.async_list_keyspaces()
            return "nonwaited_ks" not in ks_list

        await await_until_true(
            poll_interval=KEYSPACE_POLL_SLEEP_TIME,
            max_seconds=KEYSPACE_TIMEOUT,
            acondition=_awaiter2,
        )

        # check nss after dropping two of them
        keyspaces1b = set(await db_admin.async_list_keyspaces())
        assert keyspaces1b == keyspaces1

        async def _awaiter3() -> bool:
            a_info = await db_admin.async_info()
            return a_info.status == "ACTIVE"

        # drop db and check. We wait a little due to "nontransactional cluster md"
        await await_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            acondition=_awaiter3,
        )
        await db_admin.async_drop()

        db_ids = {db.id for db in (await admin.async_list_databases())}
        assert created_db_id not in db_ids

    @pytest.mark.parametrize(
        "admin_env_token", admin_test_envs_tokens(), ids=ADMIN_ENV_LIST
    )
    @pytest.mark.describe(
        "test of the full tour with AstraDBAdmin and client methods, async"
    )
    async def test_astra_db_admin_async(self, admin_env_token: tuple[str, str]) -> None:
        """
        Test plan (it has to be a single giant test to use the two DBs throughout):
        - create client -> get_admin
        - create two dbs (wait, nonwait)
        - list the two database ids, check
        - get check info on one such db through admin
        - with the client:
            4 get_dbs (a/sync, by id+region/api_endpoint), check if equal
            and test their list_collections
        - get_db_admin from the admin for one of the dbs
            - create ns
            - get_database -> list_collection_names
            - get_async_database and check == with above
        - drop dbs, (wait, nonwait)
        """
        admin_env, token = admin_env_token
        db_name_w = f"test_database_w_{admin_env}"
        db_name_nw = f"test_database_nw_{admin_env}"
        db_provider = ADMIN_ENV_VARIABLE_MAP[admin_env]["provider"]
        db_region = ADMIN_ENV_VARIABLE_MAP[admin_env]["region"]

        if db_provider is None:
            raise ValueError("Cannot proceed with provider unset.")
        if db_region is None:
            raise ValueError("Cannot proceed with region unset.")

        # create client and get admin
        client: DataAPIClient
        if admin_env == "prod":
            client = DataAPIClient(token, api_options=TESTING_SAFE_CLIENT_OPTIONS)
        else:
            client = DataAPIClient(
                token, environment=admin_env, api_options=TESTING_SAFE_CLIENT_OPTIONS
            )
        admin = client.get_admin()

        # create the two dbs
        db_admin_nw = await admin.async_create_database(
            name=db_name_nw,
            wait_until_active=False,
            cloud_provider=db_provider,
            region=db_region,
        )
        created_db_id_nw = db_admin_nw.id
        db_admin_w = await admin.async_create_database(
            name=db_name_w,
            wait_until_active=True,
            cloud_provider=db_provider,
            region=db_region,
        )
        created_db_id_w = db_admin_w.id

        async def _awaiter1() -> bool:
            db_ids = {db.id for db in (await admin.async_list_databases())}
            return created_db_id_nw in db_ids

        await await_until_true(
            poll_interval=DATABASE_POLL_SLEEP_TIME,
            max_seconds=DATABASE_TIMEOUT,
            acondition=_awaiter1,
        )

        # list, check ids
        db_ids = {db.id for db in (await admin.async_list_databases())}
        assert {created_db_id_nw, created_db_id_w} - db_ids == set()

        # get info through admin
        db_w_info = await admin.async_database_info(created_db_id_w)
        assert db_w_info.id == created_db_id_w

        # get and compare dbs obtained by the client
        synthetic_api_endpoint = API_ENDPOINT_TEMPLATE_ENV_MAP[admin_env].format(
            database_id=created_db_id_w,
            region=db_region,
        )
        adb_w_e = client.get_async_database(synthetic_api_endpoint)
        db_w_e = client.get_database(synthetic_api_endpoint)
        assert isinstance(await adb_w_e.list_collection_names(), list)
        assert db_w_e.to_async() == adb_w_e

        # get db admin from the admin and use it
        db_w_admin = admin.get_database_admin(created_db_id_w, region=db_region)
        await db_w_admin.async_create_keyspace("additional_keyspace")
        adb_w_from_admin = db_w_admin.get_async_database()
        assert isinstance(await adb_w_from_admin.list_collection_names(), list)
        db_w_from_admin = db_w_admin.get_database()
        assert db_w_from_admin.to_async() == adb_w_from_admin

        # drop databases: the w one through the admin, the nw using its db-admin
        #   (this covers most cases if combined with the
        #   (w, using db-admin) of test_astra_db_database_admin)
        assert db_admin_nw == admin.get_database_admin(
            created_db_id_nw,
            region=db_region,
        )
        # drop db and check. We wait a little due to "nontransactional cluster md"

        async def _awaiter2() -> bool:
            a_info = await db_admin_nw.async_info()
            return a_info.status == "ACTIVE"

        async def _awaiter3() -> bool:
            a_info = await db_admin_w.async_info()
            return a_info.status == "ACTIVE"

        await await_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            acondition=_awaiter2,
        )
        await await_until_true(
            poll_interval=PRE_DROP_SAFETY_POLL_INTERVAL,
            max_seconds=PRE_DROP_SAFETY_TIMEOUT,
            acondition=_awaiter3,
        )
        await db_admin_nw.async_drop(wait_until_active=False)
        await admin.async_drop_database(created_db_id_w)

        async def _awaiter4() -> bool:
            db_ids = {db.id for db in (await admin.async_list_databases())}
            return created_db_id_nw not in db_ids

        await await_until_true(
            poll_interval=DATABASE_POLL_SLEEP_TIME,
            max_seconds=DATABASE_TIMEOUT,
            acondition=_awaiter4,
        )

    @pytest.mark.describe(
        "test of the update_db_keyspace flag for AstraDBDatabaseAdmin, sync"
    )
    def test_astra_updatedbkeyspace_sync(self, sync_database: Database) -> None:
        NEW_KS_NAME_NOT_UPDATED = "tnudn_notupd"
        NEW_KS_NAME_UPDATED = "tnudn_upd"

        keyspace0 = sync_database.keyspace
        database_admin = sync_database.get_database_admin()
        database_admin.create_keyspace(NEW_KS_NAME_NOT_UPDATED)
        assert sync_database.keyspace == keyspace0

        database_admin.create_keyspace(
            NEW_KS_NAME_UPDATED,
            update_db_keyspace=True,
        )
        assert sync_database.keyspace == NEW_KS_NAME_UPDATED

        database_admin.drop_keyspace(NEW_KS_NAME_NOT_UPDATED)
        database_admin.drop_keyspace(NEW_KS_NAME_UPDATED)

    @pytest.mark.describe(
        "test of the update_db_keyspace flag for AstraDBDatabaseAdmin, async"
    )
    async def test_astra_updatedbkeyspace_async(
        self, async_database: AsyncDatabase
    ) -> None:
        NEW_KS_NAME_NOT_UPDATED = "tnudn_notupd"
        NEW_KS_NAME_UPDATED = "tnudn_upd"

        keyspace0 = async_database.keyspace
        database_admin = async_database.get_database_admin()
        await database_admin.async_create_keyspace(NEW_KS_NAME_NOT_UPDATED)
        assert async_database.keyspace == keyspace0

        await database_admin.async_create_keyspace(
            NEW_KS_NAME_UPDATED,
            update_db_keyspace=True,
        )
        assert async_database.keyspace == NEW_KS_NAME_UPDATED

        await database_admin.async_drop_keyspace(NEW_KS_NAME_NOT_UPDATED)
        await database_admin.async_drop_keyspace(NEW_KS_NAME_UPDATED)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the event-observing API
"""

from __future__ import annotations

import json
from typing import Any

import httpx
import pytest
from pytest_httpserver import HTTPServer

from astrapy import DataAPIClient
from astrapy.admin.admin import (
    async_fetch_raw_database_info_from_id_token,
    fetch_raw_database_info_from_id_token,
)
from astrapy.api_options import APIOptions, DevOpsAPIURLOptions
from astrapy.event_observers import (
    ObservableError,
    ObservableEvent,
    ObservableEventType,
    ObservableRequest,
    ObservableResponse,
    ObservableWarning,
    Observer,
    event_collector,
)
from astrapy.exceptions import DataAPIResponseException
from astrapy.exceptions.error_descriptors import (
    DataAPIErrorDescriptor,
    DataAPIWarningDescriptor,
)
from astrapy.settings.defaults import DEV_OPS_RESPONSE_HTTP_CREATED
from astrapy.utils.request_tools import HttpMethod


class TestObservingEvents:
    @pytest.mark.describe("test of regular class attached event observers, sync")
    def test_regularclass_attached_eventobservers_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Attachment test, i.e. that each of the (sync) requesting classes
        hooks to observers. Admin excluded. Sync classes.
        """
        root_endpoint = httpserver.url_for("/")

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(
            ev_list, event_types=[ObservableEventType.REQUEST]
        )
        api_options = APIOptions(event_observers={"test": my_obs})

        client = DataAPIClient(environment="other", api_options=api_options)
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        collection = database.get_collection("xcollt")
        table = database.get_table("xtablet")

        expected_c_url = "/v1/xkeyspace/xcollt"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": None}, "status": {}})
        collection.find_one()

        assert ev_list != []
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findOne": {}}

        expected_t_url = "/v1/xkeyspace/xtablet"
        httpserver.expect_oneshot_request(
            expected_t_url,
            method=HttpMethod.POST,
        ).respond_with_json(
            {"data": {"document": None}, "status": {"projectionSchema": {"x": "text"}}}
        )
        table.find_one()

        assert len(ev_list) == 2
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findOne": {}}

        expected_d_url = "/v1/xkeyspace"
        httpserver.expect_oneshot_request(
            expected_d_url,
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"tables": ["x"]}})
        database.list_table_names()

        assert len(ev_list) == 3
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"listTables": {}}

        expected_dc_url = "/v1"
        httpserver.expect_oneshot_request(
            expected_dc_url,
            method=HttpMethod.POST,
        ).respond_with_json({"a": 1})
        database.command({"z": -1}, keyspace=None)

        assert len(ev_list) == 4
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"z": -1}

    @pytest.mark.describe("test of regular class attached event observers, async")
    async def test_regularclass_attached_eventobservers_async(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Attachment test, i.e. that each of the (sync) requesting classes
        hooks to observers. Admin excluded. Async classes.
        """
        root_endpoint = httpserver.url_for("/")

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(
            ev_list, event_types=[ObservableEventType.REQUEST]
        )
        api_options = APIOptions(event_observers={"test": my_obs})

        client = DataAPIClient(environment="other", api_options=api_options)
        adatabase = client.get_async_database(root_endpoint, keyspace="xkeyspace")
        acollection = adatabase.get_collection("xcollt")
        atable = adatabase.get_table("xtablet")

        expected_c_url = "/v1/xkeyspace/xcollt"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": None}, "status": {}})
        await acollection.find_one()

        assert ev_list != []
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findOne": {}}

        expected_t_url = "/v1/xkeyspace/xtablet"
        httpserver.expect_oneshot_request(
            expected_t_url,
            method=HttpMethod.POST,
        ).respond_with_json(
            {"data": {"document": None}, "status": {"projectionSchema": {"x": "text"}}}
        )
        await atable.find_one()

        assert len(ev_list) == 2
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findOne": {}}

        expected_d_url = "/v1/xkeyspace"
        httpserver.expect_oneshot_request(
            expected_d_url,
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"tables": ["x"]}})
        await adatabase.list_table_names()

        assert len(ev_list) == 3
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"listTables": {}}

        expected_dc_url = "/v1"
        httpserver.expect_oneshot_request(
            expected_dc_url,
            method=HttpMethod.POST,
        ).respond_with_json({"a": 1})
        await adatabase.command({"z": -1}, keyspace=None)

        assert len(ev_list) == 4
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"z": -1}

    @pytest.mark.describe("test of event types being emitted, sync")
    def test_eventobservers_event_types_emitted(self, httpserver: HTTPServer) -> None:
        """
        Test about all kinds of events being emitted properly.
        Uses just a collection.
        """
        root_endpoint = httpserver.url_for("/")

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(ev_list)
        api_options = APIOptions(event_observers={"test": my_obs})

        client = DataAPIClient(environment="other", api_options=api_options)
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        collection = database.get_collection("xcollt")
        expected_url = "/v1/xkeyspace/xcollt"

        response_dict = {
            "data": {"document": None},
            "status": {
                "warnings": [
                    {"title": "Warning!"},
                ],
            },
            "errors": [
                {"title": "Error!"},
            ],
        }

        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json(response_dict)
        with pytest.raises(DataAPIResponseException):
            collection.find_one()

        assert len(ev_list) == 4
        assert isinstance(ev_list[0], ObservableRequest)
        assert json.loads(ev_list[0].payload or "") == {"findOne": {}}
        assert isinstance(ev_list[1], ObservableResponse)
        assert json.loads(ev_list[1].body or "") == response_dict
        assert isinstance(ev_list[2], ObservableWarning)
        assert ev_list[2].warning == DataAPIWarningDescriptor({"title": "Warning!"})
        assert isinstance(ev_list[3], ObservableError)
        assert ev_list[3].error == DataAPIErrorDescriptor({"title": "Error!"})

    @pytest.mark.describe("test of admin classes attached event observers, sync")
    def test_adminclasses_attached_eventobservers_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Attachment test, i.e. that each of the (sync) requesting classes
        hooks to observers. Admin classes. Sync calls.

        Note: it's a single class with a single commander: no need to test async calls.
        """
        root_endpoint = httpserver.url_for("/")
        db_id = "00000000-0000-0000-0000-000000000000"

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(
            ev_list, event_types=[ObservableEventType.REQUEST]
        )

        # Astra DB admin classes require tweaking the devops API options
        dbadmin_apioptions = APIOptions(
            event_observers={"test": my_obs},
            dev_ops_api_url_options=DevOpsAPIURLOptions(
                dev_ops_api_version="v2",
                dev_ops_url=root_endpoint,
            ),
        )
        astra_client = DataAPIClient(api_options=dbadmin_apioptions)
        astra_admin = astra_client.get_admin()
        astra_db_admin = astra_admin.get_database_admin(
            f"https://{db_id}-reg.apps.astra.datastax.com"
        )

        # AstraDBAdmin's main commander
        expected_c_url = "/v2/databases"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.GET,
        ).respond_with_json(
            [
                {
                    "id": "012",
                    "status": "bla",
                    "info": {"name": "blo", "datacenters": [], "cloudProvider": "cp"},
                    "orgId": "234",
                    "ownerId": "678",
                }
            ]
        )
        astra_admin.list_databases()

        assert ev_list != []
        assert isinstance(ev_list[-1], ObservableRequest)
        assert ev_list[-1].payload is None

        # AstraDBAdmin's region-specific commander
        # this uses another APICommander --> another test
        expected_c_url = "/v2/regions/serverless"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.GET,
        ).respond_with_json([])
        astra_admin.find_available_regions()

        assert len(ev_list) == 2
        assert isinstance(ev_list[-1], ObservableRequest)
        assert ev_list[-1].payload is None

        # AstraDBDatabaseAdmin's Data API commander
        # this will badly fail because the URL must be an Astra URL for this class
        with pytest.raises(httpx.ConnectError):
            astra_db_admin.find_embedding_providers()

        assert len(ev_list) == 3
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findEmbeddingProviders": {}}

        # AstraDBDatabaseAdmin's DevOps API commander
        expected_c_url = f"/v2/databases/{db_id}/keyspaces/xnewkeyspace"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.POST,
        ).respond_with_data(status=DEV_OPS_RESPONSE_HTTP_CREATED)
        astra_db_admin.create_keyspace("xnewkeyspace", wait_until_active=False)

        assert len(ev_list) == 4
        assert isinstance(ev_list[-1], ObservableRequest)
        assert ev_list[-1].payload is None

        # DataAPIDatabaseAdmin's (only) commander
        dapi_client = DataAPIClient(environment="other", api_options=dbadmin_apioptions)
        dapi_db_admin = dapi_client.get_database(root_endpoint).get_database_admin()
        expected_c_url = "/v1"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"keyspaces": []}})
        dapi_db_admin.list_keyspaces()

        assert len(ev_list) == 5
        assert isinstance(ev_list[-1], ObservableRequest)
        assert json.loads(ev_list[-1].payload or "") == {"findKeyspaces": {}}

    @pytest.mark.describe("test of admin functions attached event observers, sync")
    def test_adminfunctions_attached_eventobservers_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Attachment test, i.e. that each of the requesting admin functions
        hook to observers. Sync calls.
        """
        root_endpoint = httpserver.url_for("/")
        db_id = "012"

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(
            ev_list, event_types=[ObservableEventType.REQUEST]
        )

        dbadmin_apioptions = APIOptions(
            event_observers={"test": my_obs},
            dev_ops_api_url_options=DevOpsAPIURLOptions(
                dev_ops_api_version="v2",
                dev_ops_url=root_endpoint,
            ),
        )

        expected_c_url = f"/v2/databases/{db_id}"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.GET,
        ).respond_with_json({})
        fetch_raw_database_info_from_id_token(db_id, api_options=dbadmin_apioptions)

        assert ev_list != []
        assert isinstance(ev_list[-1], ObservableRequest)
        assert ev_list[-1].payload is None

    @pytest.mark.describe("test of admin functions attached event observers, async")
    async def test_adminfunctions_attached_eventobservers_async(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Attachment test, i.e. that each of the requesting admin functions
        hook to observers. Sync calls.
        """
        root_endpoint = httpserver.url_for("/")
        db_id = "012"

        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(
            ev_list, event_types=[ObservableEventType.REQUEST]
        )

        dbadmin_apioptions = APIOptions(
            event_observers={"test": my_obs},
            dev_ops_api_url_options=DevOpsAPIURLOptions(
                dev_ops_api_version="v2",
                dev_ops_url=root_endpoint,
            ),
        )

        expected_c_url = f"/v2/databases/{db_id}"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.GET,
        ).respond_with_json({})
        await async_fetch_raw_database_info_from_id_token(
            db_id, api_options=dbadmin_apioptions
        )

        assert ev_list != []
        assert isinstance(ev_list[-1], ObservableRequest)
        assert ev_list[-1].payload is None

    @pytest.mark.describe(
        "test of metadata coming with an emitted event, sampled, sync"
    )
    def test_eventobservers_eventmetadata_sampled_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Testing the metadata attached to an emitted event (sender, method name etc).
        This is just sampled in one case (a method of the sync collection).
        """
        root_endpoint = httpserver.url_for("/")

        recv_list: list[dict[str, Any]] = []

        class MyRichObserver(Observer):
            def __init__(
                self,
                rich_list: list[dict[str, Any]],
            ) -> None:
                self.evt_list = rich_list

            def receive(
                self,
                event: ObservableEvent,
                sender: Any = None,
                function_name: str | None = None,
                request_id: str | None = None,
            ) -> None:
                if event.event_type in {
                    ObservableEventType.REQUEST,
                    ObservableEventType.RESPONSE,
                }:
                    self.evt_list += [
                        {
                            "event": event,
                            "sender": sender,
                            "function_name": function_name,
                            "request_id": request_id,
                        }
                    ]

        my_r_obs = MyRichObserver(recv_list)
        api_options = APIOptions(
            event_observers={"test": my_r_obs},
        )

        client = DataAPIClient(environment="other", api_options=api_options)
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        collection = database.get_collection("xcollt")
        expected_url = "/v1/xkeyspace/xcollt"
        response_dict = {
            "data": {"document": None},
            "status": {
                "warnings": [
                    {"title": "Warning!"},
                ],
            },
            "errors": [
                {"title": "Error!"},
            ],
        }

        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json(response_dict)
        with pytest.raises(DataAPIResponseException):
            collection.find_one()

        assert len(recv_list) == 2

        rq_evt = recv_list[0]
        assert rq_evt["sender"] == collection
        assert rq_evt["function_name"] == "find_one"
        assert isinstance(rq_evt["event"], ObservableRequest)
        assert rq_evt["event"].http_method == HttpMethod.POST
        assert rq_evt["event"].url == root_endpoint.rstrip("/") + expected_url
        assert rq_evt["event"].query_parameters == {}
        expected_header_keys = {"Content-Type", "Accept", "User-Agent", "Connection"}
        found_header_keys = (rq_evt["event"].redacted_headers or {}).keys()
        assert found_header_keys - expected_header_keys == set()

        rs_evt = recv_list[1]
        assert rs_evt["sender"] == collection
        assert rs_evt["function_name"] == "find_one"
        assert isinstance(rs_evt["event"], ObservableResponse)
        assert rs_evt["event"].status_code == 200

    @pytest.mark.describe("test of request_id for emitted events, sampled, sync")
    def test_eventobservers_requestid_sampled_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Testing the request_id attached to emitted events. These must be
        the same for all events from a request, and differ between requests.
        This is sampled (a method of the sync collection + one admin "raw-req" call).
        """
        root_endpoint = httpserver.url_for("/")
        db_id = "00000000-0000-0000-0000-000000000000"

        req_ids: list[str | None] = []

        class MyReqIDObserver(Observer):
            def __init__(
                self,
                id_list: list[str | None],
            ) -> None:
                self.id_list = id_list

            def receive(
                self,
                event: ObservableEvent,
                sender: Any = None,
                function_name: str | None = None,
                request_id: str | None = None,
            ) -> None:
                self.id_list += [request_id]

        my_id_obs = MyReqIDObserver(req_ids)

        api_options = APIOptions(
            event_observers={"test_ids": my_id_obs},
        )
        client = DataAPIClient(environment="other", api_options=api_options)
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        collection = database.get_collection("xcollt")

        astra_api_options = APIOptions(
            event_observers={"test_ids": my_id_obs},
            dev_ops_api_url_options=DevOpsAPIURLOptions(
                dev_ops_api_version="v2",
                dev_ops_url=root_endpoint,
            ),
        )
        astra_client = DataAPIClient(api_options=astra_api_options)
        astra_admin = astra_client.get_admin()
        astra_db_admin = astra_admin.get_database_admin(
            f"https://{db_id}-reg.apps.astra.datastax.com"
        )

        expected_url = "/v1/xkeyspace/xcollt"
        response_dict = {
            "data": {"document": None},
            "status": {
                "warnings": [
                    {"title": "Warning!"},
                ],
            },
            "errors": [
                {"title": "Error!"},
            ],
        }
        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json(response_dict)
        with pytest.raises(DataAPIResponseException):
            collection.find_one()

        response_dict = {
            "data": {"document": None},
        }
        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json(response_dict)
        collection.find_one()

        expected_c_url = f"/v2/databases/{db_id}/keyspaces/xnewkeyspace"
        httpserver.expect_oneshot_request(
            expected_c_url,
            method=HttpMethod.POST,
        ).respond_with_data(status=DEV_OPS_RESPONSE_HTTP_CREATED)
        astra_db_admin.create_keyspace("xnewkeyspace", wait_until_active=False)

        assert len(req_ids) == 8
        assert len(set(req_ids[:4])) == 1
        assert len(set(req_ids[4:6])) == 1
        assert len(set(req_ids[6:8])) == 1
        assert len(set(req_ids)) == 3

    @pytest.mark.describe("test of event_collector utility into list, sync")
    def test_eventobservers_eventcollector_list_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Testing the event_collector context manager utility.
        """
        root_endpoint = httpserver.url_for("/")

        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")

        recv_events: list[ObservableEvent] = []
        with event_collector(
            database,
            destination=recv_events,
            event_types=[ObservableEventType.REQUEST],
        ) as inst_database:
            expected_url = "/v1/xkeyspace"
            httpserver.expect_oneshot_request(
                expected_url,
                method=HttpMethod.POST,
            ).respond_with_json({"status": {"collections": []}})
            inst_database.list_collection_names()
            inst_collection = inst_database.get_collection("xcoll")
            expected_url = "/v1/xkeyspace/xcoll"
            httpserver.expect_oneshot_request(
                expected_url,
                method=HttpMethod.POST,
            ).respond_with_json({"data": {"document": None}})
            inst_collection.find_one()

        assert len(recv_events) == 2

        # test of event collection stopping out of the 'with' block
        expected_url = "/v1/xkeyspace/xcoll"
        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": None}})
        inst_collection.find_one()

        assert len(recv_events) == 2

    @pytest.mark.describe("test of event_collector utility into dict, sync")
    def test_eventobservers_eventcollector_dict_sync(
        self, httpserver: HTTPServer
    ) -> None:
        """
        Testing the event_collector context manager utility.
        """
        root_endpoint = httpserver.url_for("/")

        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")

        recv_events: dict[ObservableEventType, list[ObservableEvent]] = {}
        with event_collector(
            database,
            destination=recv_events,
            event_types=[ObservableEventType.REQUEST],
        ) as inst_database:
            expected_url = "/v1/xkeyspace"
            httpserver.expect_oneshot_request(
                expected_url,
                method=HttpMethod.POST,
            ).respond_with_json({"status": {"collections": []}})
            inst_database.list_collection_names()
            inst_collection = inst_database.get_collection("xcoll")
            expected_url = "/v1/xkeyspace/xcoll"
            httpserver.expect_oneshot_request(
                expected_url,
                method=HttpMethod.POST,
            ).respond_with_json({"data": {"document": None}})
            inst_collection.find_one()

        assert len(recv_events) == 1
        req_evts = recv_events.get(ObservableEventType.REQUEST)
        assert req_evts is not None
        assert len(req_evts) == 2

        # test of event collection stopping out of the 'with' block
        expected_url = "/v1/xkeyspace/xcoll"
        httpserver.expect_oneshot_request(
            expected_url,
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": None}})
        inst_collection.find_one()

        assert len(recv_events) == 1
        req_evts = recv_events.get(ObservableEventType.REQUEST)
        assert req_evts is not None
        assert len(req_evts) == 2



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.data.table import (
    map2tuple_checker_insert_many,
    map2tuple_checker_insert_one,
    map2tuple_checker_update_one,
)
from astrapy.data.utils.table_converters import preprocess_table_payload
from astrapy.data_types import DataAPIMap
from astrapy.utils.api_options import SerdesOptions, defaultSerdesOptions

MAP2TUPLE_NEVER_OPTIONS = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="NEVER")
)
MAP2TUPLE_ALWAYS_OPTIONS = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="ALWAYS")
)
MAP2TUPLE_DATAAPIMAPS_OPTIONS = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="DATAAPIMAPS")
)


class TestTPreprocessorsMapsAsTuples:
    @pytest.mark.describe("test of tuple conversion as in insert_one, never convert")
    def test_map2tuple_conversion_insertone_never(self) -> None:
        payload_d = {"insertOne": {"document": {"a": {1: "x"}}}}
        expected_d = {"insertOne": {"document": {"a": {1: "x"}}}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_d == converted_d

        payload_m = {"insertOne": {"document": {"a": DataAPIMap([(1, "x")])}}}
        expected_m = {"insertOne": {"document": {"a": {1: "x"}}}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        expected_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_de == converted_de

        payload_me = {"insertOne": {"document": {"a": DataAPIMap([])}}}  # type: ignore[var-annotated]
        expected_me = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_me == converted_me

    @pytest.mark.describe("test of tuple conversion as in insert_many, never convert")
    def test_map2tuple_conversion_insertmany_never(self) -> None:
        payload_d = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        expected_d = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_d == converted_d

        payload_m = {"insertMany": {"documents": [{"a": DataAPIMap([(1, "x")])}]}}
        expected_m = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        expected_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_de == converted_de

        payload_me = {"insertMany": {"documents": [{"a": DataAPIMap([])}]}}  # type: ignore[var-annotated]
        expected_me = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_me == converted_me

    @pytest.mark.describe("test of tuple conversion as in update_one, never convert")
    def test_map2tuple_conversion_updateone_never(self) -> None:
        payload_d = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": {2: "y"}},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        expected_d = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": {2: "y"}},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_d == converted_d

        payload_m = {
            "updateOne": {
                "filter": {"f": DataAPIMap([(1, "x")])},
                "update": {
                    "$set": {"s": DataAPIMap([(2, "y")])},
                    "$unset": {"u": DataAPIMap([(3, "z")])},
                },
            }
        }
        expected_m = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": {2: "y"}},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        expected_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_de == converted_de

        payload_me = {
            "updateOne": {
                "filter": {"f": DataAPIMap([])},
                "update": {
                    "$set": {"s": DataAPIMap([])},
                    "$unset": {"u": DataAPIMap([])},
                },
            }
        }
        expected_me = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_NEVER_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_me == converted_me

    @pytest.mark.describe("test of tuple conversion as in insert_one, convert always")
    def test_map2tuple_conversion_insertone_always(self) -> None:
        payload_d = {"insertOne": {"document": {"a": {1: "x"}}}}
        expected_d = {"insertOne": {"document": {"a": [[1, "x"]]}}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_d == converted_d

        payload_m = {"insertOne": {"document": {"a": DataAPIMap([(1, "x")])}}}
        expected_m = {"insertOne": {"document": {"a": [[1, "x"]]}}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        expected_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_de == converted_de

        payload_me = {"insertOne": {"document": {"a": DataAPIMap([])}}}  # type: ignore[var-annotated]
        expected_me = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_me == converted_me

    @pytest.mark.describe("test of tuple conversion as in insert_many, convert always")
    def test_map2tuple_conversion_insertmany_always(self) -> None:
        payload_d = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        expected_d = {"insertMany": {"documents": [{"a": [[1, "x"]]}]}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_d == converted_d

        payload_m = {"insertMany": {"documents": [{"a": DataAPIMap([(1, "x")])}]}}
        expected_m = {"insertMany": {"documents": [{"a": [[1, "x"]]}]}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        expected_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_de == converted_de

        payload_me = {"insertMany": {"documents": [{"a": DataAPIMap([])}]}}  # type: ignore[var-annotated]
        expected_me = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_me == converted_me

    @pytest.mark.describe("test of tuple conversion as in update_one, convert always")
    def test_map2tuple_conversion_updateone_always(self) -> None:
        payload_d = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": {2: "y"}},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        expected_d = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": [[2, "y"]]},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_d == converted_d

        payload_m = {
            "updateOne": {
                "filter": {"f": DataAPIMap([(1, "x")])},
                "update": {
                    "$set": {"s": DataAPIMap([(2, "y")])},
                    "$unset": {"u": DataAPIMap([(3, "z")])},
                },
            }
        }
        expected_m = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": [[2, "y"]]},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        expected_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_de == converted_de

        payload_me = {
            "updateOne": {
                "filter": {"f": DataAPIMap([])},
                "update": {
                    "$set": {"s": DataAPIMap([])},
                    "$unset": {"u": DataAPIMap([])},
                },
            }
        }
        expected_me = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_ALWAYS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_me == converted_me

    @pytest.mark.describe(
        "test of tuple conversion as in insert_one, convert DataAPIMaps"
    )
    def test_map2tuple_conversion_insertone_dataapimaps(self) -> None:
        payload_d = {"insertOne": {"document": {"a": {1: "x"}}}}
        expected_d = {"insertOne": {"document": {"a": {1: "x"}}}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_d == converted_d

        payload_m = {"insertOne": {"document": {"a": DataAPIMap([(1, "x")])}}}
        expected_m = {"insertOne": {"document": {"a": [[1, "x"]]}}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        expected_de = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_de == converted_de

        payload_me = {"insertOne": {"document": {"a": DataAPIMap([])}}}  # type: ignore[var-annotated]
        expected_me = {"insertOne": {"document": {"a": {}}}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_one,
        )
        assert expected_me == converted_me

    @pytest.mark.describe(
        "test of tuple conversion as in insert_many, convert DataAPIMaps"
    )
    def test_map2tuple_conversion_insertmany_dataapimaps(self) -> None:
        payload_d = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        expected_d = {"insertMany": {"documents": [{"a": {1: "x"}}]}}
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_d == converted_d

        payload_m = {"insertMany": {"documents": [{"a": DataAPIMap([(1, "x")])}]}}
        expected_m = {"insertMany": {"documents": [{"a": [[1, "x"]]}]}}
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        expected_de = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_de == converted_de

        payload_me = {"insertMany": {"documents": [{"a": DataAPIMap([])}]}}  # type: ignore[var-annotated]
        expected_me = {"insertMany": {"documents": [{"a": {}}]}}  # type: ignore[var-annotated]
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_insert_many,
        )
        assert expected_me == converted_me

    @pytest.mark.describe(
        "test of tuple conversion as in update_one, convert DataAPIMaps"
    )
    def test_map2tuple_conversion_updateone_dataapimaps(self) -> None:
        payload_d = {
            "updateOne": {
                "filter": {"f": {1: "g"}},
                "update": {
                    "$set": {"s": {10: "t"}},
                    "$unset": {"u": {10: "v"}},
                },
            }
        }
        expected_d = {
            "updateOne": {
                "filter": {"f": {1: "g"}},
                "update": {
                    "$set": {"s": {10: "t"}},
                    "$unset": {"u": {10: "v"}},
                },
            }
        }
        converted_d = preprocess_table_payload(
            payload_d,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_d == converted_d

        payload_m = {
            "updateOne": {
                "filter": {"f": DataAPIMap([(1, "x")])},
                "update": {
                    "$set": {"s": DataAPIMap([(2, "y")])},
                    "$unset": {"u": DataAPIMap([(3, "z")])},
                },
            }
        }
        expected_m = {
            "updateOne": {
                "filter": {"f": {1: "x"}},
                "update": {
                    "$set": {"s": [[2, "y"]]},
                    "$unset": {"u": {3: "z"}},
                },
            }
        }
        converted_m = preprocess_table_payload(
            payload_m,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_m == converted_m

        # empty 'maps' must always encode as {}
        payload_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        expected_de = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_de = preprocess_table_payload(
            payload_de,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_de == converted_de

        payload_me = {
            "updateOne": {
                "filter": {"f": DataAPIMap([])},
                "update": {
                    "$set": {"s": DataAPIMap([])},
                    "$unset": {"u": DataAPIMap([])},
                },
            }
        }
        expected_me = {
            "updateOne": {
                "filter": {"f": {}},
                "update": {
                    "$set": {"s": {}},
                    "$unset": {"u": {}},
                },
            }
        }
        converted_me = preprocess_table_payload(
            payload_me,
            MAP2TUPLE_DATAAPIMAPS_OPTIONS,
            map2tuple_checker=map2tuple_checker_update_one,
        )
        assert expected_me == converted_me
